@@ -17,23 +17,35 @@
 
 Define the interfaces that are implemented by various buildbot classes.
 """
-# E0211: Method has no argument
-# E0213: Method should have "self" as first argument
-# pylint: disable-msg=E0211,E0213
 
-from zope.interface import Interface, Attribute
+# disable pylint warnings triggered by interface definitions
+# pylint: disable=no-self-argument
+# pylint: disable=no-method-argument
+
+from zope.interface import Attribute
+from zope.interface import Interface
 
 # exceptions that can be raised while trying to start a build
+
+
 class NoSlaveError(Exception):
     pass
+
+
 class BuilderInUseError(Exception):
     pass
+
+
 class BuildSlaveTooOldError(Exception):
     pass
+
+
 class LatentBuildSlaveFailedToSubstantiate(Exception):
     pass
 
+
 class IChangeSource(Interface):
+
     """
     Service which feeds Change objects to the changemaster. When files or
     directories are changed in version control, this object should represent
@@ -44,12 +56,14 @@ class IChangeSource(Interface):
     See 'Writing Change Sources' in the manual for more information.
     """
     master = Attribute('master',
-            'Pointer to BuildMaster, automatically set when started.')
+                       'Pointer to BuildMaster, automatically set when started.')
 
     def describe():
         """Return a string which briefly describes this source."""
 
+
 class ISourceStamp(Interface):
+
     """
     @cvar branch: branch from which source was drawn
     @type branch: string or None
@@ -92,18 +106,24 @@ class ISourceStamp(Interface):
         available, the caller should join them together with spaces before
         presenting them to the user."""
 
+
 class IEmailSender(Interface):
+
     """I know how to send email, and can be used by other parts of the
     Buildbot to contact developers."""
     pass
 
+
 class IEmailLookup(Interface):
+
     def getAddress(user):
         """Turn a User-name string into a valid email address. Either return
         a string (with an @ in it), None (to indicate that the user cannot
         be reached by email), or a Deferred which will fire with the same."""
 
+
 class IStatus(Interface):
+
     """I am an object, obtainable from the buildmaster, which can provide
     status information."""
 
@@ -199,27 +219,33 @@ class IStatus(Interface):
         """Unregister an IStatusReceiver. No further status messgaes will be
         delivered."""
 
+
 class IBuildSetStatus(Interface):
+
     """I represent a set of Builds, each run on a separate Builder but all
     using the same source tree."""
 
     def getReason():
         pass
+
     def getID():
         """Return the BuildSet's ID string, if any. The 'try' feature uses a
         random string as a BuildSetID to relate submitted jobs with the
         resulting BuildSet."""
     def getResponsibleUsers():
-        pass # not implemented
+        pass  # not implemented
+
     def getInterestedUsers():
-        pass # not implemented
+        pass  # not implemented
+
     def getBuilderNames():
         """Return a list of the names of all Builders on which this set will
         do builds.
-        
+
         @returns: list of names via Deferred"""
     def isFinished():
         pass
+
     def waitUntilFinished():
         """Return a Deferred that fires (with this IBuildSetStatus object)
         when all builds have finished."""
@@ -229,6 +255,7 @@ class IBuildSetStatus(Interface):
 
 
 class IBuildRequestStatus(Interface):
+
     """I represent a request to build a particular set of source code on a
     particular Builder. These requests may be merged by the time they are
     finally turned into a Build."""
@@ -267,6 +294,7 @@ class IBuildRequestStatus(Interface):
 
 
 class ISlaveStatus(Interface):
+
     def getName():
         """Return the name of the build slave."""
 
@@ -283,7 +311,9 @@ class ISlaveStatus(Interface):
         """Return a timestamp (seconds since epoch) indicating when the most
         recent message was received from the buildslave."""
 
+
 class ISchedulerStatus(Interface):
+
     def getName():
         """Return the name of this Scheduler (a string)."""
 
@@ -294,6 +324,7 @@ class ISchedulerStatus(Interface):
 
 
 class IBuilderStatus(Interface):
+
     def getName():
         """Return the name of this Builder (a string)."""
 
@@ -403,8 +434,10 @@ class IBuilderStatus(Interface):
         """Unregister an IStatusReceiver. No further status messgaes will be
         delivered."""
 
+
 class IEventSource(Interface):
-    def eventGenerator(branches=[], categories=[], committers=[], minTime=0):
+
+    def eventGenerator(branches=[], categories=[], committers=[], projects=[], minTime=0):
         """This function creates a generator which will yield all of this
         object's status events, starting with the most recent and progressing
         backwards in time. These events provide the IStatusEvent interface.
@@ -429,7 +462,9 @@ class IEventSource(Interface):
         this timestamp.
         """
 
+
 class IBuildStatus(Interface):
+
     """I represent the status of a single Build/BuildRequest. It could be
     in-progress or finished."""
 
@@ -556,7 +591,9 @@ class IBuildStatus(Interface):
         """Unregister an IStatusReceiver. No further status messgaes will be
         delivered."""
 
+
 class ITestResult(Interface):
+
     """I describe the results of a single unit test."""
 
     def getName():
@@ -581,6 +618,7 @@ class ITestResult(Interface):
 
 
 class IBuildStepStatus(Interface):
+
     """I hold status for a single BuildStep."""
 
     def getName():
@@ -621,7 +659,6 @@ class IBuildStepStatus(Interface):
         finished, this list may be incomplete (asking again later may give
         you more of them)."""
 
-
     def isFinished():
         """Return a boolean. True means the step has finished, False means it
         is still running."""
@@ -669,7 +706,9 @@ class IBuildStepStatus(Interface):
         """Unregister an IStatusReceiver. No further status messgaes will be
         delivered."""
 
+
 class IStatusEvent(Interface):
+
     """I represent a Builder Event, something non-Build related that can
     happen to a Builder."""
 
@@ -691,7 +730,9 @@ LOG_CHANNEL_STDOUT = 0
 LOG_CHANNEL_STDERR = 1
 LOG_CHANNEL_HEADER = 2
 
+
 class IStatusLog(Interface):
+
     """I represent a single Log, which is a growing list of text items that
     contains some kind of output for a single BuildStep. I might be finished,
     in which case this list has stopped growing.
@@ -724,7 +765,6 @@ class IStatusLog(Interface):
     compile log, or listing all the test cases that are going to be run).
     Such synthetic Logs are usually finished as soon as they are created."""
 
-
     def getName():
         """Returns a short string with the name of this log, probably 'log'.
         """
@@ -810,7 +850,9 @@ class IStatusLog(Interface):
         0 for stdout, 1 for stderr, 2 for header. (note that stderr is merged
         into stdout if PTYs are in use)."""
 
+
 class IStatusLogConsumer(Interface):
+
     """I am an object which can be passed to IStatusLog.subscribeConsumer().
     I represent a target for writing the contents of an IStatusLog. This
     differs from a regular IStatusReceiver in that it can pause the producer.
@@ -840,7 +882,9 @@ class IStatusLogConsumer(Interface):
     def finish():
         """The log has finished sending chunks to the consumer."""
 
+
 class IStatusReceiver(Interface):
+
     """I am an object which can receive build status updates. I may be
     subscribed to an IStatus, an IBuilderStatus, or an IBuildStatus."""
 
@@ -986,13 +1030,16 @@ class IStatusReceiver(Interface):
 
 
 class IControl(Interface):
+
     def addChange(change):
         """Add a change to the change queue, for analysis by schedulers."""
 
     def getBuilder(name):
         """Retrieve the IBuilderControl object for the given Builder."""
 
+
 class IBuilderControl(Interface):
+
     def submitBuildRequest(ss, reason, props=None):
         """Create a BuildRequest, which will eventually cause a build of the
         given SourceStamp to be run on this builder. This returns a
@@ -1030,7 +1077,9 @@ class IBuilderControl(Interface):
         # or something. However the event that is emitted is most useful in
         # the Builder column, so it kinda fits here too.
 
+
 class IBuildRequestControl(Interface):
+
     def subscribe(observer):
         """Register a callable that will be invoked (with a single
         IBuildControl object) for each Build that is created to satisfy this
@@ -1045,28 +1094,37 @@ class IBuildRequestControl(Interface):
         """Remove the build from the pending queue. Has no effect if the
         build has already been started."""
 
+
 class IBuildControl(Interface):
+
     def getStatus():
         """Return an IBuildStatus object for the Build that I control."""
     def stopBuild(reason="<no reason given>"):
         """Halt the build. This has no effect if the build has already
         finished."""
 
+
 class ILogFile(Interface):
+
     """This is the internal interface to a LogFile, used by the BuildStep to
     write data into the log.
     """
     def addStdout(data):
         pass
+
     def addStderr(data):
         pass
+
     def addHeader(data):
         pass
+
     def finish():
         """The process that is feeding the log file has finished, and no
         further data will be added. This closes the logfile."""
 
+
 class ILogObserver(Interface):
+
     """Objects which provide this interface can be used in a BuildStep to
     watch the output of a LogFile and parse it incrementally.
     """
@@ -1074,6 +1132,7 @@ class ILogObserver(Interface):
     # internal methods
     def setStep(step):
         pass
+
     def setLog(log):
         pass
 
@@ -1081,11 +1140,14 @@ class ILogObserver(Interface):
     def logChunk(build, step, log, channel, text):
         pass
 
+
 class IBuildSlave(Interface):
     # this is a marker interface for the BuildSlave class
     pass
 
+
 class ILatentBuildSlave(IBuildSlave):
+
     """A build slave that is not always running, but can run when requested.
     """
     substantiated = Attribute('Substantiated',
@@ -1114,7 +1176,9 @@ class ILatentBuildSlave(IBuildSlave):
         build finished.
         """
 
+
 class IRenderable(Interface):
+
     """An object that can be interpolated with properties from a build.
     """
 
@@ -1123,7 +1187,10 @@ class IRenderable(Interface):
 
         @param iprops: the L{IProperties} provider supplying the properties.
         """
+
+
 class IProperties(Interface):
+
     """
     An object providing access to build properties
     """
@@ -1198,10 +1265,13 @@ class IProperties(Interface):
         @returns: rendered value
         """
 
+
 class IScheduler(Interface):
     pass
 
+
 class ITriggerableScheduler(Interface):
+
     """
     A scheduler that can be triggered by buildsteps.
     """
@@ -1210,7 +1280,8 @@ class ITriggerableScheduler(Interface):
         """Trigger a build with the given source stamp and properties.
         """
 
+
 class IBuildStepFactory(Interface):
+
     def buildStep():
-        """
-        """
+        pass
