@@ -16,18 +16,25 @@
 
 import re
 import types
-from email.Message import Message
-from email.Utils import formatdate
-from email.MIMEText import MIMEText
-from email.MIMENonMultipart import MIMENonMultipart
-from email.MIMEMultipart import MIMEMultipart
-from StringIO import StringIO
 import urllib
 
-from zope.interface import implements
-from twisted.internet import defer, reactor
-from twisted.mail.smtp import ESMTPSenderFactory
+from StringIO import StringIO
+from email.message import Message
+from email.mime.multipart import MIMEMultipart
+from email.mime.nonmultipart import MIMENonMultipart
+from email.mime.text import MIMEText
+from email.utils import formatdate
+
+from twisted.internet import defer
+from twisted.internet import reactor
 from twisted.python import log as twlog
+from zope.interface import implements
+
+try:
+    from twisted.mail.smtp import ESMTPSenderFactory
+    ESMTPSenderFactory = ESMTPSenderFactory  # for pyflakes
+except ImportError:
+    ESMTPSenderFactory = None
 
 have_ssl = True
 try:
@@ -36,14 +43,40 @@ try:
 except ImportError:
     have_ssl = False
 
-from buildbot import interfaces, util
+# this incantation teaches email to output utf-8 using 7- or 8-bit encoding,
+# although it has no effect before python-2.7.
+from email import charset
+charset.add_charset('utf-8', charset.SHORTEST, None, 'utf-8')
+
+from buildbot import config
+from buildbot import interfaces
+from buildbot import util
 from buildbot.process.users import users
 from buildbot.status import base
-from buildbot.status.results import FAILURE, SUCCESS, Results
+from buildbot.status import buildset
+from buildbot.status.results import EXCEPTION
+from buildbot.status.results import FAILURE
+from buildbot.status.results import Results
+from buildbot.status.results import SUCCESS
+from buildbot.status.results import WARNINGS
 
-VALID_EMAIL = re.compile("[a-zA-Z0-9\.\_\%\-\+]+@[a-zA-Z0-9\.\_\%\-]+.[a-zA-Z]{2,6}")
+# Email parsing can be complex. We try to take a very liberal
+# approach. The local part of an email address matches ANY non
+# whitespace character. Rather allow a malformed email address than
+# croaking on a valid (the matching of domains should be correct
+# though; requiring the domain to not be a top level domain). With
+# these regular expressions, we can match the following:
+#
+#    full.name@example.net
+#    Full Name <full.name@example.net>
+#    <full.name@example.net>
+VALID_EMAIL_ADDR = r"(?:\S+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)+\.?)"
+VALID_EMAIL = re.compile(r"^(?:%s|(.+\s+)?<%s>\s*)$" % ((VALID_EMAIL_ADDR,) * 2))
+VALID_EMAIL_ADDR = re.compile(VALID_EMAIL_ADDR)
 
 ENCODING = 'utf8'
+LOG_ENCODING = 'utf-8'
+
 
 class Domain(util.ComparableMixin):
     implements(interfaces.IEmailLookup)
@@ -60,76 +93,134 @@ class Domain(util.ComparableMixin):
         return name + "@" + self.domain
 
 
-def defaultMessage(mode, name, build, results, master_status):
-    """Generate a buildbot mail message and return a tuple of message text
-        and type."""
-    result = Results[results]
-    ss = build.getSourceStamp()
+def _defaultMessageIntro(mode, results, build):
+    prev = build.getPreviousBuild()
 
+    if results == FAILURE:
+        if "change" in mode and prev and prev.getResults() != results or \
+                "problem" in mode and prev and prev.getResults() != FAILURE:
+            text = "The Buildbot has detected a new failure"
+        else:
+            text = "The Buildbot has detected a failed build"
+    elif results == WARNINGS:
+        text = "The Buildbot has detected a problem in the build"
+    elif results == SUCCESS:
+        if "change" in mode and prev and prev.getResults() != results:
+            text = "The Buildbot has detected a restored build"
+        else:
+            text = "The Buildbot has detected a passing build"
+    elif results == EXCEPTION:
+        text = "The Buildbot has detected a build exception"
+
+    return text
+
+
+def _defaultMessageProjects(source_stamps, master_status):
+    projects = []
+
+    for ss in source_stamps:
+        if ss.project and ss.project not in projects:
+            projects.append(ss.project)
+
+    if not projects:
+        projects = [master_status.getTitle()]
+
+    return ', '.join(projects)
+
+
+def _defaultMessageURLs(master_status, build):
     text = ""
-    if mode == "all":
-        text += "The Buildbot has finished a build"
-    elif mode == "failing":
-        text += "The Buildbot has detected a failed build"
-    elif mode == "warnings":
-        text += "The Buildbot has detected a problem in the build"
-    elif mode == "passing":
-        text += "The Buildbot has detected a passing build"
-    elif mode == "change" and result == 'success':
-        text += "The Buildbot has detected a restored build"
-    else:    
-        text += "The Buildbot has detected a new failure"
-    if ss and ss.project:
-        project = ss.project
-    else:
-        project = master_status.getTitle()
-    text += " on builder %s while building %s.\n" % (name, project)
-    if master_status.getURLForThing(build):
-        text += "Full details are available at:\n %s\n" % master_status.getURLForThing(build)
-    text += "\n"
 
-    if master_status.getBuildbotURL():
-        text += "Buildbot URL: %s\n\n" % urllib.quote(master_status.getBuildbotURL(), '/:')
-
-    text += "Buildslave for this Build: %s\n\n" % build.getSlavename()
-    text += "Build Reason: %s\n" % build.getReason()
-
-    source = ""
-    if ss and ss.branch:
-        source += "[branch %s] " % ss.branch
-    if ss and ss.revision:
-        source += str(ss.revision)
-    else:
-        source += "HEAD"
-    if ss and ss.patch:
-        source += " (plus patch)"
-
-    text += "Build Source Stamp: %s\n" % source
-
-    text += "Blamelist: %s\n" % ",".join(build.getResponsibleUsers())
+    # add build URL if available
+    url = master_status.getURLForThing(build)
+    if url:
+        text += " Full details are available at:\n    %s\n" % url
 
     text += "\n"
 
+    # add buildbot main URL if available
+    url = master_status.getBuildbotURL()
+    if url:
+        text += "Buildbot URL: %s\n\n" % urllib.quote(url, '/:')
+
+    return text
+
+
+def _defaultMessageSourceStamps(source_stamps):
+    text = ""
+
+    for ss in source_stamps:
+        source = ""
+
+        if ss.branch:
+            source += "[branch %s] " % ss.branch
+
+        if ss.revision:
+            source += str(ss.revision)
+        else:
+            source += "HEAD"
+
+        if ss.patch:
+            source += " (plus patch)"
+
+        discriminator = ""
+        if ss.codebase:
+            discriminator = " '%s'" % ss.codebase
+
+        text += "Build Source Stamp%s: %s\n" % (discriminator, source)
+
+    return text
+
+
+def _defaultMessageSummary(build, results):
     t = build.getText()
     if t:
         t = ": " + " ".join(t)
     else:
         t = ""
 
-    if result == 'success':
-        text += "Build succeeded!\n"
-    elif result == 'warnings':
-        text += "Build Had Warnings%s\n" % t
+    if results == SUCCESS:
+        text = "Build succeeded!\n"
+    elif results == WARNINGS:
+        text = "Build Had Warnings%s\n" % t
     else:
-        text += "BUILD FAILED%s\n" % t
+        text = "BUILD FAILED%s\n" % t
 
+    return text
+
+
+def defaultMessage(mode, name, build, results, master_status):
+    """Generate a buildbot mail message and return a tuple of message text
+        and type."""
+    ss_list = build.getSourceStamps()
+
+    text = _defaultMessageIntro(mode, results, build)
+    text += " on builder %s while building %s." % \
+            (name, _defaultMessageProjects(ss_list, master_status))
+
+    text += _defaultMessageURLs(master_status, build)
+
+    text += "Buildslave for this Build: %s\n\n" % build.getSlavename()
+    text += "Build Reason: %s\n" % build.getReason()
+    text += _defaultMessageSourceStamps(ss_list)
+    text += "Blamelist: %s\n" % ",".join(build.getResponsibleUsers())
     text += "\n"
-    text += "sincerely,\n"
+
+    text += _defaultMessageSummary(build, results)
+    text += "\n"
+    text += "Sincerely,\n"
     text += " -The Buildbot\n"
     text += "\n"
-    return { 'body' : text, 'type' : 'plain' }
 
-class MailNotifier(base.StatusReceiverMultiService):
+    return {'body': text, 'type': 'plain'}
+
+
+def defaultGetPreviousBuild(current_build):
+    return current_build.getPreviousBuild()
+
+
+class MailNotifier(base.StatusReceiverMultiService, buildset.BuildSetSummaryNotifierMixin):
+
     """This is a status notifier which sends email to a list of recipients
     upon the completion of each build. It can be configured to only send out
     mail for certain builds, and only send messages when the build fails, or
@@ -155,16 +246,18 @@ class MailNotifier(base.StatusReceiverMultiService):
                      "subject", "sendToInterestedUsers", "customMesg",
                      "messageFormatter", "extraHeaders"]
 
-    possible_modes = ('all', 'failing', 'problem', 'change', 'passing', 'warnings')
+    possible_modes = ("change", "failing", "passing", "problem", "warnings", "exception")
 
-    def __init__(self, fromaddr, mode="all", categories=None, builders=None,
-                 addLogs=False, relayhost="localhost", buildSetSummary=False,
+    def __init__(self, fromaddr, mode=("failing", "passing", "warnings"),
+                 categories=None, builders=None, addLogs=False,
+                 relayhost="localhost", buildSetSummary=False,
                  subject="buildbot %(result)s in %(title)s on %(builder)s",
                  lookup=None, extraRecipients=[],
                  sendToInterestedUsers=True, customMesg=None,
                  messageFormatter=defaultMessage, extraHeaders=None,
-                 addPatch=True, useTls=False, 
-                 smtpUser=None, smtpPassword=None, smtpPort=25):
+                 addPatch=True, useTls=False,
+                 smtpUser=None, smtpPassword=None, smtpPort=25,
+                 previousBuildGetter=defaultGetPreviousBuild):
         """
         @type  fromaddr: string
         @param fromaddr: the email address to be used in the 'From' header.
@@ -173,7 +266,7 @@ class MailNotifier(base.StatusReceiverMultiService):
                                       of the Interested Users. If False, only
                                       send mail to the extraRecipients list.
 
-        @type  extraRecipients: tuple of string
+        @type  extraRecipients: tuple of strings
         @param extraRecipients: a list of email addresses to which messages
                                 should be sent (in addition to the
                                 InterestedUsers list, which includes any
@@ -189,15 +282,17 @@ class MailNotifier(base.StatusReceiverMultiService):
                         %(builder)s will be replaced with the name of the
                         builder which provoked the message.
 
-        @type  mode: string (defaults to all)
-        @param mode: one of MailNotifer.possible_modes:
-                     - 'all': send mail about all builds, passing and failing
-                     - 'failing': only send mail about builds which fail
-                     - 'warnings': send mail if builds contain warnings or fail 
-                     - 'passing': only send mail about builds which succeed
-                     - 'problem': only send mail about a build which failed
-                     when the previous build passed
-                     - 'change': only send mail about builds who change status
+        @type  mode: list of strings
+        @param mode: a list of MailNotifer.possible_modes:
+                     - "change":  send mail about builds which change status
+                     - "failing": send mail about builds which fail
+                     - "passing": send mail about builds which succeed
+                     - "problem": send mail about a build which failed
+                                  when the previous build passed
+                     - "warnings": send mail if a build contain warnings
+                     - "exception": send mail if a build fails due to an exception
+                     - "all": always send mail
+                     Defaults to ("failing", "passing", "warnings").
 
         @type  builders: list of strings
         @param builders: a list of builder names for which mail should be
@@ -223,12 +318,12 @@ class MailNotifier(base.StatusReceiverMultiService):
         @type  relayhost: string
         @param relayhost: the host to which the outbound SMTP connection
                           should be made. Defaults to 'localhost'
-                          
+
         @type  buildSetSummary: boolean
         @param buildSetSummary: if True, this notifier will only send a summary
                                 email when a buildset containing any of its
                                 watched builds completes
-                                
+
         @type  lookup:    implementor of {IEmailLookup}
         @param lookup:    object which provides IEmailLookup, which is
                           responsible for mapping User names for Interested
@@ -264,7 +359,7 @@ class MailNotifier(base.StatusReceiverMultiService):
                              values may be WithProperties instances.
 
         @type useTls: boolean
-        @param useTls: Send emails using TLS and authenticate with the 
+        @param useTls: Send emails using TLS and authenticate with the
                        smtp host. Defaults to False.
 
         @type smtpUser: string
@@ -278,33 +373,61 @@ class MailNotifier(base.StatusReceiverMultiService):
         @type smtpPort: int
         @param smtpPort: The port that will be used when connecting to the
                          relayhost. Defaults to 25.
-        """
 
+        @type previousBuildGetter: func
+        @param previousBuildGetter: function taking a BuildStatus instance
+                                    returning a BuildStatus of the build
+                                    previous to the one passed in. This allows
+                                    to implement a relative ordering between
+                                    builds other than the default one, which is
+                                    chronological.
+        """
         base.StatusReceiverMultiService.__init__(self)
-        assert isinstance(extraRecipients, (list, tuple))
-        for r in extraRecipients:
-            assert isinstance(r, str)
-            # require full email addresses, not User names
-            assert VALID_EMAIL.search(r), "%s is not a valid email" % r 
+
+        if not isinstance(extraRecipients, (list, tuple)):
+            config.error("extraRecipients must be a list or tuple")
+        else:
+            for r in extraRecipients:
+                if not isinstance(r, str) or not VALID_EMAIL.search(r):
+                    config.error(
+                        "extra recipient %r is not a valid email" % (r,))
         self.extraRecipients = extraRecipients
         self.sendToInterestedUsers = sendToInterestedUsers
         self.fromaddr = fromaddr
-        assert mode in self.possible_modes
+        if isinstance(mode, basestring):
+            if mode == "all":
+                mode = ("failing", "passing", "warnings", "exception")
+            elif mode == "warnings":
+                mode = ("failing", "warnings")
+            else:
+                mode = (mode,)
+        for m in mode:
+            if m not in self.possible_modes:
+                if m == "all":
+                    config.error(
+                        "mode 'all' is not valid in an iterator and must be passed in as a separate string")
+                else:
+                    config.error(
+                        "mode %s is not a valid mode" % (m,))
         self.mode = mode
         self.categories = categories
         self.builders = builders
         self.addLogs = addLogs
         self.relayhost = relayhost
+        if '\n' in subject:
+            config.error(
+                'Newlines are not allowed in email subjects')
         self.subject = subject
         if lookup is not None:
-            if type(lookup) is str:
+            if isinstance(lookup, str):
                 lookup = Domain(lookup)
             assert interfaces.IEmailLookup.providedBy(lookup)
         self.lookup = lookup
         self.customMesg = customMesg
         self.messageFormatter = messageFormatter
         if extraHeaders:
-            assert isinstance(extraHeaders, dict)
+            if not isinstance(extraHeaders, dict):
+                config.error("extraHeaders must be a dictionary")
         self.extraHeaders = extraHeaders
         self.addPatch = addPatch
         self.useTls = useTls
@@ -313,96 +436,100 @@ class MailNotifier(base.StatusReceiverMultiService):
         self.smtpPort = smtpPort
         self.buildSetSummary = buildSetSummary
         self.buildSetSubscription = None
+        self.getPreviousBuild = previousBuildGetter
         self.watched = []
         self.master_status = None
 
         # you should either limit on builders or categories, not both
-        if self.builders != None and self.categories != None:
-            twlog.err("Please specify only builders or categories to include not both.")
-            raise interfaces.ParameterError("Please specify only builders or categories to include not both.")
+        if self.builders is not None and self.categories is not None:
+            config.error(
+                "Please specify only builders or categories to include - " +
+                "not both.")
 
         if customMesg:
-            twlog.msg("customMesg is deprecated; please use messageFormatter instead")
+            config.error(
+                "customMesg is deprecated; use messageFormatter instead")
 
     def setServiceParent(self, parent):
         """
         @type  parent: L{buildbot.master.BuildMaster}
         """
         base.StatusReceiverMultiService.setServiceParent(self, parent)
-        self.setup()
-
-    def setup(self):
-        self.master_status = self.parent.getStatus()
+        self.master_status = self.parent
         self.master_status.subscribe(self)
-        
-            
+        self.master = self.master_status.master
+
     def startService(self):
         if self.buildSetSummary:
-            self.buildSetSubscription = \
-            self.parent.subscribeToBuildsetCompletions(self.buildsetFinished)
- 
+            self.summarySubscribe()
+
         base.StatusReceiverMultiService.startService(self)
-        
-   
+
     def stopService(self):
-        if self.buildSetSubscription is not None:
-            self.buildSetSubscription.unsubscribe()
-            self.buildSetSubscription = None
-            
+        self.summaryUnsubscribe()
+
         return base.StatusReceiverMultiService.stopService(self)
 
     def disownServiceParent(self):
         self.master_status.unsubscribe(self)
+        self.master_status = None
         for w in self.watched:
             w.unsubscribe(self)
         return base.StatusReceiverMultiService.disownServiceParent(self)
 
     def builderAdded(self, name, builder):
         # only subscribe to builders we are interested in
-        if self.categories != None and builder.category not in self.categories:
+        if self.categories is not None and builder.category not in self.categories:
             return None
 
         self.watched.append(builder)
-        return self # subscribe to this builder
+        return self  # subscribe to this builder
 
     def builderRemoved(self, name):
         pass
 
     def builderChangedState(self, name, state):
         pass
+
     def buildStarted(self, name, build):
         pass
+
     def isMailNeeded(self, build, results):
         # here is where we actually do something.
         builder = build.getBuilder()
         if self.builders is not None and builder.name not in self.builders:
-            return False # ignore this build
+            return False  # ignore this build
         if self.categories is not None and \
-               builder.category not in self.categories:
-            return False # ignore this build
+                builder.category not in self.categories:
+            return False  # ignore this build
 
-        if self.mode == "warnings" and results == SUCCESS:
-            return False
-        if self.mode == "failing" and results != FAILURE:
-            return False
-        if self.mode == "passing" and results != SUCCESS:
-            return False
-        if self.mode == "problem":
-            if results != FAILURE:
-                return False
-            prev = build.getPreviousBuild()
-            if prev and prev.getResults() == FAILURE:
-                return False
-        if self.mode == "change":
-            prev = build.getPreviousBuild()
-            if not prev or prev.getResults() == results:
-                return False
-        
-        return True
+        prev = self.getPreviousBuild(build)
+        if "change" in self.mode:
+            if prev and prev.getResults() != results:
+                return True
+        if "failing" in self.mode and results == FAILURE:
+            return True
+        if "passing" in self.mode and results == SUCCESS:
+            return True
+        if "problem" in self.mode and results == FAILURE:
+            if prev and prev.getResults() != FAILURE:
+                return True
+        if "warnings" in self.mode and results == WARNINGS:
+            return True
+        if "exception" in self.mode and results == EXCEPTION:
+            return True
+
+        return False
+
+    def sendBuildSetSummary(self, buildset, builds):
+        # only include builds for which isMailNeeded returns true
+        builds = [build for build in builds if self.isMailNeeded(build, build.getResults())]
+        if builds:
+            self.buildMessage("(whole buildset)", builds, buildset['results'])
 
     def buildFinished(self, name, build, results):
-        if ( not self.buildSetSummary and
-             self.isMailNeeded(build, results) ):
+        if (not self.buildSetSummary and
+                self.isMailNeeded(build, results)):
             # for testing purposes, buildMessage returns a Deferred that fires
             # when the mail has been sent. To help unit tests, we return that
             # Deferred here even though the normal IStatusReceiver.buildFinished
@@ -411,40 +538,6 @@ class MailNotifier(base.StatusReceiverMultiService):
             # rearrange this.
             return self.buildMessage(name, [build], results)
         return None
-    
-    def _gotBuilds(self, res, builddicts, buildset, builders):
-        builds = []
-        for (builddictlist, builder) in zip(builddicts, builders):
-                for builddict in builddictlist:
-                    build = builder.getBuild(builddict['number'])
-                    if self.isMailNeeded(build, build.results):
-                        builds.append(build)
-
-        self.buildMessage("Buildset Complete: " + buildset['reason'], builds,
-                          buildset['results'])
-        
-    def _gotBuildRequests(self, breqs, buildset):
-        builddicts = []
-        builders =[]
-        dl = []
-        for breq in breqs:
-            buildername = breq['buildername']
-            builders.append(self.master_status.getBuilder(buildername))
-            d = self.parent.db.builds.getBuildsForRequest(breq['brid'])
-            d.addCallback(builddicts.append)
-            dl.append(d)
-        d = defer.DeferredList(dl)
-        d.addCallback(self._gotBuilds, builddicts, buildset, builders)
-
-    def _gotBuildSet(self, buildset, bsid):
-        d = self.parent.db.buildrequests.getBuildRequests(bsid=bsid)
-        d.addCallback(self._gotBuildRequests, buildset)
-        
-    def buildsetFinished(self, bsid, result):
-        d = self.parent.db.buildsets.getBuildset(bsid=bsid)
-        d.addCallback(self._gotBuildSet, bsid)
-            
-        return d
 
     def getCustomMesgData(self, mode, name, build, results, master_status):
         #
@@ -473,7 +566,7 @@ class MailNotifier(base.StatusReceiverMultiService):
                  'buildText': build.getText(),
                  'buildProperties': build.getProperties(),
                  'slavename': build.getSlavename(),
-                 'reason':  build.getReason(),
+                 'reason': build.getReason().replace('\n', ''),
                  'responsibleUsers': build.getResponsibleUsers(),
                  'branch': "",
                  'revision': "",
@@ -482,13 +575,27 @@ class MailNotifier(base.StatusReceiverMultiService):
                  'changes': [],
                  'logs': logs}
 
-        ss = build.getSourceStamp()
-        if ss:
-            attrs['branch'] = ss.branch
-            attrs['revision'] = ss.revision
-            attrs['patch'] = ss.patch
-            attrs['patch_info'] = ss.patch_info
-            attrs['changes'] = ss.changes[:]
+        ss = None
+        ss_list = build.getSourceStamps()
+
+        if ss_list:
+            if len(ss_list) == 1:
+                ss = ss_list[0]
+                if ss:
+                    attrs['branch'] = ss.branch
+                    attrs['revision'] = ss.revision
+                    attrs['patch'] = ss.patch
+                    attrs['patch_info'] = ss.patch_info
+                    attrs['changes'] = ss.changes[:]
+            else:
+                for key in ['branch', 'revision', 'patch', 'patch_info', 'changes']:
+                    attrs[key] = {}
+                for ss in ss_list:
+                    attrs['branch'][ss.codebase] = ss.branch
+                    attrs['revision'][ss.codebase] = ss.revision
+                    attrs['patch'][ss.codebase] = ss.patch
+                    attrs['patch_info'][ss.codebase] = ss.patch_info
+                    attrs['changes'][ss.codebase] = ss.changes[:]
 
         return attrs
 
@@ -497,9 +604,9 @@ class MailNotifier(base.StatusReceiverMultiService):
         # we'll attach it as MIMEText; otherwise, it gets attached as a binary
         # file.  This will suit the vast majority of cases, since utf8 is by
         # far the most common encoding.
-        if type(patch[1]) != types.UnicodeType:
+        if not isinstance(patch[1], types.UnicodeType):
             try:
-                    unicode = patch[1].decode('utf8')
+                unicode = patch[1].decode('utf8')
             except UnicodeDecodeError:
                 unicode = None
         else:
@@ -512,7 +619,7 @@ class MailNotifier(base.StatusReceiverMultiService):
             a = MIMENonMultipart('application', 'octet-stream')
             a.set_payload(patch[1])
         a.add_header('Content-Disposition', "attachment",
-                    filename="source patch " + str(index) )
+                     filename="source patch " + str(index))
         return a
 
     def createEmail(self, msgdict, builderName, title, results, builds=None,
@@ -522,12 +629,14 @@ class MailNotifier(base.StatusReceiverMultiService):
         if 'subject' in msgdict:
             subject = msgdict['subject'].encode(ENCODING)
         else:
-            subject = self.subject % { 'result': Results[results],
-                                       'projectName': title,
-                                       'title': title,
-                                       'builder': builderName,
-                                       }
+            subject = self.subject % {'result': Results[results],
+                                      'projectName': title,
+                                      'title': title,
+                                      'builder': builderName,
+                                      }
 
+        assert '\n' not in subject, \
+            "Subject cannot contain newlines"
 
         assert type in ('plain', 'html'), \
             "'%s' message type must be 'plain' or 'html'." % type
@@ -553,59 +662,77 @@ class MailNotifier(base.StatusReceiverMultiService):
             for log in logs:
                 name = "%s.%s" % (log.getStep().getName(),
                                   log.getName())
-                if ( self._shouldAttachLog(log.getName()) or
-                     self._shouldAttachLog(name) ):
-                    a = MIMEText(log.getText().encode(ENCODING), 
+                if (self._shouldAttachLog(log.getName()) or
+                        self._shouldAttachLog(name)):
+                    # Use distinct filenames for the e-mail summary
+                    if self.buildSetSummary:
+                        filename = "%s.%s.%s" % (log.getStep().getBuild().getBuilder().getName(),
+                                                 log.getStep().getName(),
+                                                 log.getName())
+                    else:
+                        filename = name
+
+                    text = log.getText()
+                    if not isinstance(text, unicode):
+                        # guess at the encoding, and use replacement symbols
+                        # for anything that's not in that encoding
+                        text = text.decode(LOG_ENCODING, 'replace')
+                    a = MIMEText(text.encode(ENCODING),
                                  _charset=ENCODING)
                     a.add_header('Content-Disposition', "attachment",
-                                 filename=name)
+                                 filename=filename)
                     m.attach(a)
 
-        #@todo: is there a better way to do this?
+        # @todo: is there a better way to do this?
         # Add any extra headers that were requested, doing WithProperties
         # interpolation if only one build was given
         if self.extraHeaders:
-            for k,v in self.extraHeaders.items():
-                if len(builds) == 1:
-                    k = interfaces.IProperties(builds[0]).render(k)
-                if k in m:
-                    twlog.msg("Warning: Got header " + k +
-                      " in self.extraHeaders "
-                      "but it already exists in the Message - "
-                      "not adding it.")
-                if len(builds) == 1:
-                    m[k] = interfaces.IProperties(builds[0]).render(v)
-                else:
+            if len(builds) == 1:
+                d = builds[0].render(self.extraHeaders)
+            else:
+                d = defer.succeed(self.extraHeaders)
+
+            @d.addCallback
+            def addExtraHeaders(extraHeaders):
+                for k, v in extraHeaders.items():
+                    if k in m:
+                        twlog.msg("Warning: Got header " + k +
+                                  " in self.extraHeaders "
+                                  "but it already exists in the Message - "
+                                  "not adding it.")
                     m[k] = v
-    
-        return m
-    
+            d.addCallback(lambda _: m)
+            return d
+
+        return defer.succeed(m)
+
     def buildMessageDict(self, name, build, results):
         if self.customMesg:
             # the customMesg stuff can be *huge*, so we prefer not to load it
             attrs = self.getCustomMesgData(self.mode, name, build, results,
                                            self.master_status)
             text, type = self.customMesg(attrs)
-            msgdict = { 'body' : text, 'type' : type }
+            msgdict = {'body': text, 'type': type}
         else:
             msgdict = self.messageFormatter(self.mode, name, build, results,
                                             self.master_status)
-        
-        return msgdict
 
+        return msgdict
 
     def buildMessage(self, name, builds, results):
         patches = []
         logs = []
-        msgdict = {"body":""}
-        
+        msgdict = {"body": ""}
+
         for build in builds:
-            ss = build.getSourceStamp()
-            if ss and ss.patch and self.addPatch:
-                patches.append(ss.patch)
+            ss_list = build.getSourceStamps()
+            if self.addPatch:
+                for ss in ss_list:
+                    if ss.patch:
+                        patches.append(ss.patch)
             if self.addLogs:
                 logs.extend(build.getLogs())
-            
+
             tmp = self.buildMessageDict(name=build.getBuilder().name,
                                         build=build, results=build.results)
             msgdict['body'] += tmp['body']
@@ -614,76 +741,47 @@ class MailNotifier(base.StatusReceiverMultiService):
             if "subject" in tmp:
                 msgdict['subject'] = tmp['subject']
 
-        m = self.createEmail(msgdict, name, self.master_status.getTitle(),
+        d = self.createEmail(msgdict, name, self.master_status.getTitle(),
                              results, builds, patches, logs)
 
-        # now, who is this message going to?
-        self.dl = []
-        self.recipients = []
-        if self.sendToInterestedUsers:
-            for build in builds:
-                d = defer.succeed(build)
-                if self.lookup:
-                    d.addCallback(self.useLookup)
-                else:
-                    d.addCallback(self.useUsers)
-        else:
-            d = defer.DeferredList(self.dl)
-        d.addCallback(self._gotRecipients, self.recipients, m)
+        @d.addCallback
+        def getRecipients(m):
+            # now, who is this message going to?
+            if self.sendToInterestedUsers:
+                dl = []
+                for build in builds:
+                    if self.lookup:
+                        d = self.useLookup(build)
+                    else:
+                        d = self.useUsers(build)
+                    dl.append(d)
+                d = defer.gatherResults(dl)
+            else:
+                d = defer.succeed([])
+            d.addCallback(self._gotRecipients, m)
         return d
 
     def useLookup(self, build):
-        for u in build.getInterestedUsers():
+        dl = []
+        for u in build.getResponsibleUsers() + build.getInterestedUsers():
             d = defer.maybeDeferred(self.lookup.getAddress, u)
-            d.addCallback(self.recipients.append)
-            self.dl.append(d)
-        return defer.DeferredList(self.dl)
+            dl.append(d)
+        return defer.gatherResults(dl)
 
     def useUsers(self, build):
-        self.contacts = []
-        ss = build.getSourceStamp()
-        for change in ss.changes:
-            d = self.parent.db.changes.getChangeUids(change.number)
-            def getContacts(uids):
-                def uidContactPair(contact, uid):
-                    return (contact, uid)
-                d = defer.succeed(None)
-                for uid in uids:
-                    d.addCallback(lambda _ :
-                                     users.getUserContact(self.parent,
-                                                          contact_type='email',
-                                                          uid=uid))
-                    d.addCallback(lambda contact: uidContactPair(contact, uid))
-                    d.addCallback(self.contacts.append)
-                return d
-            d.addCallback(getContacts)
-            def logNoMatch(_):
-                for pair in self.contacts:
-                    contact, uid = pair
-                    if contact is None:
-                        twlog.msg("Unable to find email for uid: %r" % uid)
-                return [pair[0] for pair in self.contacts]
-            d.addCallback(logNoMatch)
-            d.addCallback(self.recipients.extend)
-            def addOwners(_):
-                owners = [e for e in build.getInterestedUsers()
-                          if e not in build.getResponsibleUsers()]
-                self.recipients.extend(owners)
-            d.addCallback(addOwners)
-            self.dl.append(d)
-        return defer.DeferredList(self.dl)
+        return users.getBuildContacts(self.master, build, ['email'])
 
     def _shouldAttachLog(self, logname):
-        if type(self.addLogs) is bool:
+        if isinstance(self.addLogs, bool):
             return self.addLogs
         return logname in self.addLogs
 
-    def _gotRecipients(self, res, rlist, m):
+    def _gotRecipients(self, rlist, m):
         to_recipients = set()
         cc_recipients = set()
 
-        for r in rlist:
-            if r is None: # getAddress didn't like this address
+        for r in reduce(list.__add__, rlist, []):
+            if r is None:  # getAddress didn't like this address
                 continue
 
             # Git can give emails like 'User' <user@foo.com>@foo.com so check
@@ -694,7 +792,7 @@ class MailNotifier(base.StatusReceiverMultiService):
             if VALID_EMAIL.search(r):
                 to_recipients.add(r)
             else:
-                twlog.msg("INVALID EMAIL: %r" + r)
+                twlog.msg("INVALID EMAIL: %r" % r)
 
         # If we're sending to interested users put the extras in the
         # CC list so they can tell if they are also interested in the
@@ -712,7 +810,7 @@ class MailNotifier(base.StatusReceiverMultiService):
 
     def sendmail(self, s, recipients):
         result = defer.Deferred()
-        
+
         if have_ssl and self.useTls:
             client_factory = ssl.ClientContextFactory()
             client_factory.method = SSLv3_METHOD
@@ -723,7 +821,10 @@ class MailNotifier(base.StatusReceiverMultiService):
             useAuth = True
         else:
             useAuth = False
-        
+
+        if not ESMTPSenderFactory:
+            raise RuntimeError("twisted-mail is not installed - cannot "
+                               "send mail")
         sender_factory = ESMTPSenderFactory(
             self.smtpUser, self.smtpPassword,
             self.fromaddr, recipients, StringIO(s),
@@ -732,11 +833,10 @@ class MailNotifier(base.StatusReceiverMultiService):
             requireAuthentication=useAuth)
 
         reactor.connectTCP(self.relayhost, self.smtpPort, sender_factory)
-        
+
         return result
 
     def sendMessage(self, m, recipients):
         s = m.as_string()
         twlog.msg("sending mail (%d bytes) to" % len(s), recipients)
         return self.sendmail(s, recipients)
-
