@@ -14,52 +14,98 @@
 # Copyright Buildbot Team Members
 
 
-from twisted.web import html
-from twisted.web.util import Redirect
-import urllib, time
-from twisted.python import log
-from twisted.internet import defer
+import collections
+import time
+import urllib
+
 from buildbot import interfaces
-from buildbot.status.web.base import HtmlResource, BuildLineMixin, \
-    path_to_build, path_to_slave, path_to_builder, path_to_change, \
-    path_to_root, getAndCheckProperties, ICurrentBox, build_get_class, \
-    map_branches, path_to_authfail, ActionResource
-
-from buildbot.status.web.build import BuildsResource, StatusResourceBuild
 from buildbot import util
+from buildbot.schedulers.forcesched import ForceScheduler
+from buildbot.schedulers.forcesched import ValidationError
+from buildbot.status.web.base import ActionResource
+from buildbot.status.web.base import BuildLineMixin
+from buildbot.status.web.base import HtmlResource
+from buildbot.status.web.base import ICurrentBox
+from buildbot.status.web.base import build_get_class
+from buildbot.status.web.base import getRequestCharset
+from buildbot.status.web.base import map_branches
+from buildbot.status.web.base import path_to_authzfail
+from buildbot.status.web.base import path_to_build
+from buildbot.status.web.base import path_to_builder
+from buildbot.status.web.base import path_to_change
+from buildbot.status.web.base import path_to_root
+from buildbot.status.web.base import path_to_slave
+from buildbot.status.web.build import BuildsResource
+from buildbot.status.web.build import StatusResourceBuild
+from twisted.internet import defer
+from twisted.python import log
+from twisted.web import html
 
-class ForceAllBuildsActionResource(ActionResource):
+
+class ForceAction(ActionResource):
+
+    @defer.inlineCallbacks
+    def force(self, req, builderNames):
+        master = self.getBuildmaster(req)
+        owner = self.getAuthz(req).getUsernameFull(req)
+        schedulername = req.args.get("forcescheduler", ["<unknown>"])[0]
+        if schedulername == "<unknown>":
+            defer.returnValue((path_to_builder(req, self.builder_status),
+                               "forcescheduler arg not found"))
+            return
+
+        args = {}
+        # decode all of the args
+        encoding = getRequestCharset(req)
+        for name, argl in req.args.iteritems():
+            if name == "checkbox":
+                # damn html's ungeneric checkbox implementation...
+                for cb in argl:
+                    args[cb.decode(encoding)] = True
+            else:
+                args[name] = [arg.decode(encoding) for arg in argl]
+
+        for sch in master.allSchedulers():
+            if schedulername == sch.name:
+                try:
+                    yield sch.force(owner, builderNames, **args)
+                    msg = ""
+                except ValidationError, e:
+                    msg = html.escape(e.message.encode('ascii', 'ignore'))
+                break
+        else:
+            msg = 'no scheduler found'
+
+        # send the user back to the builder page
+        defer.returnValue(msg)
+
+
+class ForceAllBuildsActionResource(ForceAction):
 
     def __init__(self, status, selectedOrAll):
         self.status = status
         self.selectedOrAll = selectedOrAll
         self.action = "forceAllBuilds"
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         authz = self.getAuthz(req)
-        d = authz.actionAllowed('forceAllBuilds', req)
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        res = wfd.getResult()
+        res = yield authz.actionAllowed('forceAllBuilds', req)
 
         if not res:
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
             return
 
-        builders = None
         if self.selectedOrAll == 'all':
-            builders = self.status.getBuilderNames()
+            builderNames = None
         elif self.selectedOrAll == 'selected':
-            builders = [b for b in req.args.get("selected", []) if b]
+            builderNames = [b for b in req.args.get("selected", []) if b]
 
-        for bname in builders:
-            builder_status = self.status.getBuilder(bname)
-            build = StatusResourceBuilder(builder_status)
-            build.force(req, auth_ok=True) # auth_ok because we already checked
+        msg = yield self.force(req, builderNames)
 
         # back to the welcome page
-        yield path_to_root(req)
+        defer.returnValue((path_to_root(req) + "builders", msg))
+
 
 class StopAllBuildsActionResource(ActionResource):
 
@@ -68,15 +114,12 @@ class StopAllBuildsActionResource(ActionResource):
         self.selectedOrAll = selectedOrAll
         self.action = "stopAllBuilds"
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         authz = self.getAuthz(req)
-        d = authz.actionAllowed('stopAllBuilds', req)
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        res = wfd.getResult()
+        res = yield authz.actionAllowed('stopAllBuilds', req)
         if not res:
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
             return
 
         builders = None
@@ -96,8 +139,46 @@ class StopAllBuildsActionResource(ActionResource):
                     continue
                 build = StatusResourceBuild(build_status)
                 build.stop(req, auth_ok=True)
+
         # go back to the welcome page
-        yield path_to_root(req)
+        defer.returnValue(path_to_root(req))
+
+
+class CancelAllPendingBuildsActionResource(ActionResource):
+
+    def __init__(self, status, selectedOrAll):
+        self.status = status
+        self.selectedOrAll = selectedOrAll
+        self.action = 'cancelAllPendingBuilds'
+
+    @defer.inlineCallbacks
+    def performAction(self, req):
+        authz = self.getAuthz(req)
+        res = yield authz.actionAllowed('cancelAllPendingBuilds', req)
+        if not res:
+            defer.returnValue(path_to_authzfail(req))
+            return
+
+        builders = None
+        if self.selectedOrAll == 'all':
+            builders = self.status.getBuilderNames()
+        elif self.selectedOrAll == 'selected':
+            builders = [b for b in req.args.get("selected", []) if b]
+
+        c = interfaces.IControl(self.getBuildmaster(req))
+        for bname in builders:
+            authz = self.getAuthz(req)
+            builder_control = c.getBuilder(bname)
+
+            brcontrols = yield builder_control.getPendingBuildRequestControls()
+
+            for build_req in brcontrols:
+                log.msg("Cancelling %s" % build_req)
+                build_req.cancel()
+
+        # go back to the welcome page
+        defer.returnValue(path_to_root(req))
+
 
 class PingBuilderActionResource(ActionResource):
 
@@ -105,98 +186,92 @@ class PingBuilderActionResource(ActionResource):
         self.builder_status = builder_status
         self.action = "pingBuilder"
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         log.msg("web ping of builder '%s'" % self.builder_status.getName())
-        d = self.getAuthz(req).actionAllowed('pingBuilder', req, self.builder_status)
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        res = wfd.getResult()
+        res = yield self.getAuthz(req).actionAllowed('pingBuilder', req,
+                                                     self.builder_status)
         if not res:
             log.msg("..but not authorized")
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
             return
 
         c = interfaces.IControl(self.getBuildmaster(req))
         bc = c.getBuilder(self.builder_status.getName())
         bc.ping()
         # send the user back to the builder page
-        yield path_to_builder(req, self.builder_status)
+        defer.returnValue(path_to_builder(req, self.builder_status))
 
-class ForceBuildActionResource(ActionResource):
 
-    def __init__(self, builder_status, req_args):
+class ForceBuildActionResource(ForceAction):
+
+    def __init__(self, builder_status):
         self.builder_status = builder_status
-        self.req_args = req_args
         self.action = "forceBuild"
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         # check if this is allowed
-        d = self.getAuthz(req).actionAllowed(self.action, req,
-                                             self.builder_status)
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        res = wfd.getResult()
+        res = yield self.getAuthz(req).actionAllowed(self.action, req,
+                                                     self.builder_status)
         if not res:
             log.msg("..but not authorized")
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
             return
 
-        master = self.getBuildmaster(req)
+        builderName = self.builder_status.getName()
 
-        # keep weird stuff out of the branch revision, and property strings.
-        branch_validate = master.config.validation['branch']
-        revision_validate = master.config.validation['revision']
-        if not branch_validate.match(self.req_args['branch']):
-            log.msg("bad branch '%s'" % self.req_args['branch'])
-            yield path_to_builder(req, self.builder_status)
-            return
-        if not revision_validate.match(self.req_args['revision']):
-            log.msg("bad revision '%s'" % self.req_args['revision'])
-            yield path_to_builder(req, self.builder_status)
-            return
-        properties = getAndCheckProperties(req)
-        if properties is None:
-            yield path_to_builder(req, self.builder_status)
-            return
-        if not self.req_args['branch']:
-            self.req_args['branch'] = None
-        if not self.req_args['revision']:
-            self.req_args['revision'] = None
-
-        d = master.db.sourcestamps.addSourceStamp(
-                                      branch=self.req_args['branch'],
-                                      revision=self.req_args['revision'],
-                                      project=self.req_args['project'],
-                                      repository=self.req_args['repository'])
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        ssid = wfd.getResult()
-
-        r = ("The web-page 'force build' button was pressed by '%s': %s\n"
-             % (html.escape(self.req_args['name']),
-                html.escape(self.req_args['reason'])))
-        d = master.addBuildset(builderNames=[self.builder_status.getName()],
-                               ssid=ssid, reason=r,
-                               properties=properties.asDict())
-        wfd = defer.waitForDeferred(d)
-        yield wfd
-        tup = wfd.getResult()
-        # check that (bsid, brids) were properly stored
-        if not isinstance(tup, (int, dict)):
-            log.err("(ignored) while trying to force build")
+        msg = yield self.force(req, [builderName])
 
         # send the user back to the builder page
-        yield path_to_builder(req, self.builder_status)
+        defer.returnValue((path_to_builder(req, self.builder_status), msg))
+
+
+def buildForceContextForField(req, default_props, sch, field, master, buildername):
+    pname = "%s.%s" % (sch.name, field.fullName)
+
+    default = field.default
+
+    if "list" in field.type:
+        choices = field.getChoices(master, sch, buildername)
+        default_props[pname + ".choices"] = choices
+
+    default = req.args.get(pname, [default])[0]
+    if "bool" in field.type:
+        default = "checked" if default else ""
+    elif isinstance(default, unicode):
+        # filter out unicode chars, and html stuff
+        default = html.escape(default.encode('utf-8', 'ignore'))
+
+    default_props[pname] = default
+
+    if "nested" in field.type:
+        for subfield in field.fields:
+            buildForceContextForField(req, default_props, sch, subfield, master, buildername)
+
+
+def buildForceContext(cxt, req, master, buildername=None):
+    force_schedulers = {}
+    default_props = collections.defaultdict(str)
+    for sch in master.allSchedulers():
+        if isinstance(sch, ForceScheduler) and (buildername is None or(buildername in sch.builderNames)):
+            force_schedulers[sch.name] = sch
+            for field in sch.all_fields:
+                buildForceContextForField(req, default_props, sch, field, master, buildername)
+
+    cxt['force_schedulers'] = force_schedulers
+    cxt['default_props'] = default_props
 
 # /builders/$builder
+
+
 class StatusResourceBuilder(HtmlResource, BuildLineMixin):
     addSlash = True
 
-    def __init__(self, builder_status):
+    def __init__(self, builder_status, numbuilds=20):
         HtmlResource.__init__(self)
         self.builder_status = builder_status
+        self.numbuilds = numbuilds
 
     def getPageTitle(self, request):
         return "Buildbot: %s" % self.builder_status.getName()
@@ -211,7 +286,7 @@ class StatusResourceBuilder(HtmlResource, BuildLineMixin):
         if when is not None:
             b['when'] = util.formatInterval(when)
             b['when_time'] = time.strftime("%H:%M:%S",
-                                      time.localtime(time.time() + when))
+                                           time.localtime(time.time() + when))
 
         step = build.getCurrentStep()
         # TODO: is this necessarily the case?
@@ -227,54 +302,76 @@ class StatusResourceBuilder(HtmlResource, BuildLineMixin):
 
         return b
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def content(self, req, cxt):
         b = self.builder_status
 
+        # Grab all the parameters which are prefixed with 'property.'.
+        # We'll use these to filter the builds and build requests we
+        # show below.
+        props = {}
+        prop_prefix = 'property.'
+        for arg, val in req.args.iteritems():
+            if arg.startswith(prop_prefix):
+                props[arg[len(prop_prefix):]] = val[0]
+
+        def prop_match(oprops):
+            for key, val in props.iteritems():
+                if key not in oprops or val != str(oprops[key]):
+                    return False
+            return True
+
         cxt['name'] = b.getName()
+        cxt['description'] = b.getDescription()
         req.setHeader('Cache-Control', 'no-cache')
         slaves = b.getSlaves()
         connected_slaves = [s for s in slaves if s.isConnected()]
 
-        cxt['current'] = [self.builder(x, req) for x in b.getCurrentBuilds()]
+        cxt['current'] = [
+            self.builder(x, req) for x in b.getCurrentBuilds()
+            if prop_match(x.getProperties())]
 
         cxt['pending'] = []
-        wfd = defer.waitForDeferred(
-            b.getPendingBuildRequestStatuses())
-        yield wfd
-        statuses = wfd.getResult()
+        statuses = yield b.getPendingBuildRequestStatuses()
         for pb in statuses:
             changes = []
 
-            wfd = defer.waitForDeferred(
-                    pb.getSourceStamp())
-            yield wfd
-            source = wfd.getResult()
+            source = yield pb.getSourceStamp()
+            submitTime = yield pb.getSubmitTime()
+            bsid = yield pb.getBsid()
 
-            wfd = defer.waitForDeferred(
-                    pb.getSubmitTime())
-            yield wfd
-            submitTime = wfd.getResult()
+            properties = yield \
+                pb.master.db.buildsets.getBuildsetProperties(bsid)
+            if not prop_match(properties):
+                continue
 
             if source.changes:
                 for c in source.changes:
-                    changes.append({ 'url' : path_to_change(req, c),
-                                     'who' : c.who,
-                                     'revision' : c.revision,
-                                     'repo' : c.repository })
+                    changes.append({'url': path_to_change(req, c),
+                                    'who': c.who,
+                                    'revision': c.revision,
+                                    'repo': c.repository})
 
             cxt['pending'].append({
                 'when': time.strftime("%b %d %H:%M:%S",
                                       time.localtime(submitTime)),
                 'delay': util.formatInterval(util.now() - submitTime),
                 'id': pb.brid,
-                'changes' : changes,
-                'num_changes' : len(changes),
-                })
+                'changes': changes,
+                'num_changes': len(changes),
+                'properties': properties,
+            })
 
-        numbuilds = int(req.args.get('numbuilds', ['5'])[0])
+        try:
+            numbuilds = cxt['numbuilds'] = int(req.args.get('numbuilds', [self.numbuilds])[0])
+        except ValueError:
+            numbuilds = cxt['numbuilds'] = 10
+        maxsearch = int(req.args.get('maxsearch', [200])[0])
         recent = cxt['recent'] = []
-        for build in b.generateFinishedBuilds(num_builds=int(numbuilds)):
+        for build in b.generateFinishedBuilds(
+                num_builds=int(numbuilds),
+                max_search=maxsearch,
+                filter_fn=lambda b: prop_match(b.getProperties())):
             recent.append(self.get_line_values(req, build, False))
 
         sl = cxt['slaves'] = []
@@ -285,74 +382,24 @@ class StatusResourceBuilder(HtmlResource, BuildLineMixin):
             s['link'] = path_to_slave(req, slave)
             s['name'] = slave.getName()
             c = s['connected'] = slave.isConnected()
+            s['paused'] = slave.isPaused()
+            s['admin'] = slave.getAdmin() or u''
             if c:
-                s['admin'] = unicode(slave.getAdmin() or '', 'utf-8')
                 connected_slaves += 1
         cxt['connected_slaves'] = connected_slaves
 
         cxt['authz'] = self.getAuthz(req)
         cxt['builder_url'] = path_to_builder(req, b)
-
+        buildForceContext(cxt, req, self.getBuildmaster(req), b.getName())
         template = req.site.buildbot_service.templates.get_template("builder.html")
-        yield template.render(**cxt)
-
-    def force(self, req, auth_ok=False):
-        name = req.args.get("username", ["<unknown>"])[0]
-        reason = req.args.get("comments", ["<no reason specified>"])[0]
-        branch = req.args.get("branch", [""])[0]
-        revision = req.args.get("revision", [""])[0]
-        repository = req.args.get("repository", [""])[0]
-        project = req.args.get("project", [""])[0]
-
-        log.msg("web forcebuild of builder '%s', branch='%s', revision='%s',"
-                " repository='%s', project='%s' by user '%s'" % (
-                self.builder_status.getName(), branch, revision, repository,
-                project, name))
-
-        if not auth_ok:
-            req_args = dict(name=name, reason=reason, branch=branch,
-                            revision=revision, repository=repository,
-                            project=project)
-            return ForceBuildActionResource(self.builder_status, req_args)
-
-        master = self.getBuildmaster(req)
-
-        # keep weird stuff out of the branch revision, and property strings.
-        branch_validate = master.config.validation['branch']
-        revision_validate = master.config.validation['revision']
-        if not branch_validate.match(branch):
-            log.msg("bad branch '%s'" % branch)
-            return Redirect(path_to_builder(req, self.builder_status))
-        if not revision_validate.match(revision):
-            log.msg("bad revision '%s'" % revision)
-            return Redirect(path_to_builder(req, self.builder_status))
-        properties = getAndCheckProperties(req)
-        if properties is None:
-            return Redirect(path_to_builder(req, self.builder_status))
-        if not branch:
-            branch = None
-        if not revision:
-            revision = None
-
-        d = master.db.sourcestamps.addSourceStamp(branch=branch,
-                revision=revision, project=project, repository=repository)
-        def make_buildset(ssid):
-            r = ("The web-page 'force build' button was pressed by '%s': %s\n"
-                 % (html.escape(name), html.escape(reason)))
-            return master.addBuildset(
-                    builderNames=[self.builder_status.getName()],
-                    ssid=ssid, reason=r, properties=properties.asDict())
-        d.addCallback(make_buildset)
-        d.addErrback(log.err, "(ignored) while trying to force build")
-        # send the user back to the builder page
-        return Redirect(path_to_builder(req, self.builder_status))
+        defer.returnValue(template.render(**cxt))
 
     def ping(self, req):
         return PingBuilderActionResource(self.builder_status)
 
     def getChild(self, path, req):
         if path == "force":
-            return self.force(req)
+            return ForceBuildActionResource(self.builder_status)
         if path == "ping":
             return self.ping(req)
         if path == "cancelbuild":
@@ -364,13 +411,14 @@ class StatusResourceBuilder(HtmlResource, BuildLineMixin):
 
         return HtmlResource.getChild(self, path, req)
 
+
 class CancelChangeResource(ActionResource):
 
     def __init__(self, builder_status):
         ActionResource.__init__(self)
         self.builder_status = builder_status
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         try:
             request_id = req.args.get("id", [None])[0]
@@ -387,32 +435,27 @@ class CancelChangeResource(ActionResource):
             c = interfaces.IControl(self.getBuildmaster(req))
             builder_control = c.getBuilder(self.builder_status.getName())
 
-            wfd = defer.waitForDeferred(
-                    builder_control.getPendingBuildRequestControls())
-            yield wfd
-            brcontrols = wfd.getResult()
+            brcontrols = yield builder_control.getPendingBuildRequestControls()
 
             for build_req in brcontrols:
                 if cancel_all or (build_req.brid == request_id):
                     log.msg("Cancelling %s" % build_req)
-                    d = authz.actionAllowed('cancelPendingBuild',
-                                            req, build_req)
-                    wfd = defer.waitForDeferred(d)
-                    yield wfd
-                    res = wfd.getResult()
+                    res = yield authz.actionAllowed('cancelPendingBuild', req,
+                                                    build_req)
                     if res:
                         build_req.cancel()
                     else:
-                        yield path_to_authfail(req)
+                        defer.returnValue(path_to_authzfail(req))
                         return
                     if not cancel_all:
                         break
 
-        yield path_to_builder(req, self.builder_status)
+        defer.returnValue(path_to_builder(req, self.builder_status))
+
 
 class StopChangeMixin(object):
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def stopChangeForBuilder(self, req, builder_status, auth_ok=False):
         try:
             request_change = req.args.get("change", [None])[0]
@@ -425,23 +468,14 @@ class StopChangeMixin(object):
             c = interfaces.IControl(self.getBuildmaster(req))
             builder_control = c.getBuilder(builder_status.getName())
 
-            wfd = defer.waitForDeferred(
-                    builder_control.getPendingBuildRequestControls())
-            yield wfd
-            brcontrols = wfd.getResult()
-
+            brcontrols = yield builder_control.getPendingBuildRequestControls()
             build_controls = dict((x.brid, x) for x in brcontrols)
 
-            wfd = defer.waitForDeferred(
-                    builder_status.getPendingBuildRequestStatuses())
-            yield wfd
-            build_req_statuses = wfd.getResult()
+            build_req_statuses = yield \
+                builder_status.getPendingBuildRequestStatuses()
 
             for build_req in build_req_statuses:
-                wfd = defer.waitForDeferred(
-                        build_req.getSourceStamp())
-                yield wfd
-                ss = wfd.getResult()
+                ss = yield build_req.getSourceStamp()
 
                 if not ss.changes:
                     continue
@@ -450,17 +484,14 @@ class StopChangeMixin(object):
                     if change.number == request_change:
                         control = build_controls[build_req.brid]
                         log.msg("Cancelling %s" % control)
-                        d = authz.actionAllowed('stopChange', req, control)
-                        wfd = defer.waitForDeferred(d)
-                        yield wfd
-                        res = wfd.getResult()
+                        res = yield authz.actionAllowed('stopChange', req, control)
                         if (auth_ok or res):
                             control.cancel()
                         else:
-                            yield False
+                            defer.returnValue(False)
                             return
 
-        yield True
+        defer.returnValue(True)
 
 
 class StopChangeResource(StopChangeMixin, ActionResource):
@@ -469,18 +500,15 @@ class StopChangeResource(StopChangeMixin, ActionResource):
         ActionResource.__init__(self)
         self.builder_status = builder_status
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         """Cancel all pending builds that include a given numbered change."""
-        wfd = defer.waitForDeferred(
-            self.stopChangeForBuilder(req, self.builder_status))
-        yield wfd
-        success = wfd.getResult()
+        success = yield self.stopChangeForBuilder(req, self.builder_status)
 
         if not success:
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
         else:
-            yield path_to_builder(req, self.builder_status)
+            defer.returnValue(path_to_builder(req, self.builder_status))
 
 
 class StopChangeAllResource(StopChangeMixin, ActionResource):
@@ -489,27 +517,23 @@ class StopChangeAllResource(StopChangeMixin, ActionResource):
         ActionResource.__init__(self)
         self.status = status
 
-    @defer.deferredGenerator
+    @defer.inlineCallbacks
     def performAction(self, req):
         """Cancel all pending builds that include a given numbered change."""
         authz = self.getAuthz(req)
-        wfd = defer.waitForDeferred(authz.actionAllowed('stopChange', req))
-        yield wfd
-        res = wfd.getResult()
+        res = yield authz.actionAllowed('stopChange', req)
         if not res:
-            yield path_to_authfail(req)
+            defer.returnValue(path_to_authzfail(req))
             return
 
         for bname in self.status.getBuilderNames():
             builder_status = self.status.getBuilder(bname)
-            wfd = defer.waitForDeferred(
-                self.stopChangeForBuilder(req, builder_status, auth_ok=True))
-            yield wfd
-            if not wfd.getResult():
-                yield path_to_authfail(req)
+            res = yield self.stopChangeForBuilder(req, builder_status, auth_ok=True)
+            if not res:
+                defer.returnValue(path_to_authzfail(req))
                 return
 
-        yield path_to_root(req)
+        defer.returnValue(path_to_root(req))
 
 
 # /builders/_all
@@ -526,6 +550,8 @@ class StatusResourceAllBuilders(HtmlResource, BuildLineMixin):
             return self.stopall(req)
         if path == "stopchangeall":
             return StopChangeAllResource(self.status)
+        if path == "cancelpendingall":
+            return CancelAllPendingBuildsActionResource(self.status, 'all')
 
         return HtmlResource.getChild(self, path, req)
 
@@ -536,6 +562,8 @@ class StatusResourceAllBuilders(HtmlResource, BuildLineMixin):
         return StopAllBuildsActionResource(self.status, 'all')
 
 # /builders/_selected
+
+
 class StatusResourceSelectedBuilders(HtmlResource, BuildLineMixin):
 
     def __init__(self, status):
@@ -547,6 +575,8 @@ class StatusResourceSelectedBuilders(HtmlResource, BuildLineMixin):
             return self.forceselected(req)
         if path == "stopselected":
             return self.stopselected(req)
+        if path == "cancelpendingselected":
+            return CancelAllPendingBuildsActionResource(self.status, 'selected')
 
         return HtmlResource.getChild(self, path, req)
 
@@ -557,20 +587,33 @@ class StatusResourceSelectedBuilders(HtmlResource, BuildLineMixin):
         return StopAllBuildsActionResource(self.status, 'selected')
 
 # /builders
+
+
 class BuildersResource(HtmlResource):
     pageTitle = "Builders"
     addSlash = True
 
-    @defer.deferredGenerator
+    def __init__(self, numbuilds=20):
+        HtmlResource.__init__(self)
+        self.numbuilds = numbuilds
+
+    @defer.inlineCallbacks
     def content(self, req, cxt):
         status = self.getStatus(req)
+        encoding = getRequestCharset(req)
 
-        builders = req.args.get("builder", status.getBuilderNames())
-        branches = [b for b in req.args.get("branch", []) if b]
+        showCategories = req.args.get("category", [])
+        if showCategories == []:
+            showCategories = None
+        builders = req.args.get("builder", status.getBuilderNames(categories=showCategories))
+        branches = [b.decode(encoding)
+                    for b in req.args.get("branch", [])
+                    if b]
 
         # get counts of pending builds for each builder
         brstatus_ds = []
         brcounts = {}
+
         def keep_count(statuses, builderName):
             brcounts[builderName] = len(statuses)
         for builderName in builders:
@@ -578,10 +621,7 @@ class BuildersResource(HtmlResource):
             d = builder_status.getPendingBuildRequestStatuses()
             d.addCallback(keep_count, builderName)
             brstatus_ds.append(d)
-        wfd = defer.waitForDeferred(
-            defer.gatherResults(brstatus_ds))
-        yield wfd
-        wfd.getResult()
+        yield defer.gatherResults(brstatus_ds)
 
         cxt['branches'] = branches
         bs = cxt['builders'] = []
@@ -590,8 +630,8 @@ class BuildersResource(HtmlResource):
         online = 0
         base_builders_url = path_to_root(req) + "builders/"
         for bn in builders:
-            bld = { 'link': base_builders_url + urllib.quote(bn, safe=''),
-                    'name': bn }
+            bld = {'link': base_builders_url + urllib.quote(bn, safe=''),
+                   'name': bn}
             bs.append(bld)
 
             builder = status.getBuilder(bn)
@@ -600,7 +640,11 @@ class BuildersResource(HtmlResource):
             if builds:
                 b = builds[0]
                 bld['build_url'] = (bld['link'] + "/builds/%d" % b.getNumber())
-                label = b.getProperty("got_revision")
+                label = None
+                all_got_revisions = b.getAllGotRevisions()
+                # If len = 1 then try if revision can be used as label.
+                if len(all_got_revisions) == 1:
+                    label = all_got_revisions[all_got_revisions.keys()[0]]
                 if not label or len(str(label)) > 20:
                     label = "#%d" % b.getNumber()
 
@@ -621,19 +665,18 @@ class BuildersResource(HtmlResource):
         cxt['authz'] = self.getAuthz(req)
         cxt['num_building'] = building
         cxt['num_online'] = online
-
+        buildForceContext(cxt, req, self.getBuildmaster(req))
         template = req.site.buildbot_service.templates.get_template("builders.html")
-        yield template.render(**cxt)
+        defer.returnValue(template.render(**cxt))
 
     def getChild(self, path, req):
         s = self.getStatus(req)
         if path in s.getBuilderNames():
             builder_status = s.getBuilder(path)
-            return StatusResourceBuilder(builder_status)
+            return StatusResourceBuilder(builder_status, self.numbuilds)
         if path == "_all":
             return StatusResourceAllBuilders(self.getStatus(req))
         if path == "_selected":
             return StatusResourceSelectedBuilders(self.getStatus(req))
 
         return HtmlResource.getChild(self, path, req)
-
