@@ -19,15 +19,19 @@
 # but "the rest" is pretty minimal
 
 import re
-from twisted.web import resource
-from twisted.python.reflect import namedModule
-from twisted.python import log
+
 from twisted.internet import defer
+from twisted.python import log
+from twisted.python.reflect import namedModule
+from twisted.web import resource
+from twisted.web import server
+
 
 class ChangeHookResource(resource.Resource):
-     # this is a cheap sort of template thingy
+    # this is a cheap sort of template thingy
     contentType = "text/html; charset=utf-8"
-    children    = {}
+    children = {}
+
     def __init__(self, dialects={}):
         """
         The keys of 'dialects' select a modules to load under
@@ -37,75 +41,88 @@ class ChangeHookResource(resource.Resource):
         """
         self.dialects = dialects
         self.request_dialect = None
-    
+
     def getChild(self, name, request):
         return self
 
     def render_GET(self, request):
         """
-        Reponds to events and starts the build process
+        Responds to events and starts the build process
           different implementations can decide on what methods they will accept
         """
         return self.render_POST(request)
 
     def render_POST(self, request):
         """
-        Reponds to events and starts the build process
+        Responds to events and starts the build process
           different implementations can decide on what methods they will accept
-        
+
         :arguments:
             request
                 the http request object
         """
 
         try:
-            changes, src = self.getChanges( request )
-        except ValueError, err:
-            request.setResponseCode(400, err.args[0])
-            return defer.succeed(err.args[0])
+            changes, src = self.getChanges(request)
+        except ValueError, val_err:
+            request.setResponseCode(400, val_err.args[0])
+            return val_err.args[0]
+        except Exception, e:
+            log.err(e, "processing changes from web hook")
+            msg = "Error processing changes."
+            request.setResponseCode(500, msg)
+            return msg
 
         log.msg("Payload: " + str(request.args))
-        
+
         if not changes:
             log.msg("No changes found")
-            return defer.succeed("no changes found")
-        d = self.submitChanges( changes, request, src )
-        d.addCallback(lambda _ : "OK")
-        return d
+            return "no changes found"
+        d = self.submitChanges(changes, request, src)
+
+        def ok(_):
+            request.setResponseCode(202)
+            request.finish()
+
+        def err(why):
+            log.err(why, "adding changes from web hook")
+            request.setResponseCode(500)
+            request.finish()
+        d.addCallbacks(ok, err)
+        return server.NOT_DONE_YET
 
-    
     def getChanges(self, request):
         """
         Take the logic from the change hook, and then delegate it
         to the proper handler
         http://localhost/change_hook/DIALECT will load up
         buildmaster/status/web/hooks/DIALECT.py
-        
+
         and call getChanges()
-        
+
         the return value is a list of changes
-        
+
         if DIALECT is unspecified, a sample implementation is provided
         """
         uriRE = re.search(r'^/change_hook/?([a-zA-Z0-9_]*)', request.uri)
-        
+
         if not uriRE:
             log.msg("URI doesn't match change_hook regex: %s" % request.uri)
             raise ValueError("URI doesn't match change_hook regex: %s" % request.uri)
-        
+
         changes = []
         src = None
-        
+
         # Was there a dialect provided?
         if uriRE.group(1):
             dialect = uriRE.group(1)
         else:
             dialect = 'base'
-            
+
         if dialect in self.dialects.keys():
             log.msg("Attempting to load module buildbot.status.web.hooks." + dialect)
             tempModule = namedModule('buildbot.status.web.hooks.' + dialect)
-            changes, src = tempModule.getChanges(request,self.dialects[dialect])
+            changes, src = tempModule.getChanges(request, self.dialects[dialect])
             log.msg("Got the following changes %s" % changes)
             self.request_dialect = dialect
         else:
@@ -115,12 +132,10 @@ class ChangeHookResource(resource.Resource):
             raise ValueError(m)
 
         return (changes, src)
-                
-    @defer.deferredGenerator
+
+    @defer.inlineCallbacks
     def submitChanges(self, changes, request, src):
         master = request.site.buildbot_service.master
         for chdict in changes:
-            wfd = defer.waitForDeferred(master.addChange(src=src, **chdict))
-            yield wfd
-            change = wfd.getResult()
+            change = yield master.addChange(src=src, **chdict)
             log.msg("injected change %s" % change)
